#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <errno.h>
#include <strings.h>
#include <signal.h>

unsigned int port;

void cleanup();
void awaitConnections();
void sendMessage();

int main(int argc, char **argv) {

	if (argc != 2) {
		fprintf(stderr, "usage: ./socktest port \n");
		exit(1);
	}

	signal(SIGTERM, cleanup);
	signal(SIGALRM, cleanup);

	port = atoi(argv[1]);

	/* first, we start listening */
	if( fork() == 0)
 		awaitConnections();

	/* we wait for the child to start listening */
	sleep(2);

	/* send a message to ourselves, now that we're listening */
	sendMessage();

	return(0);
}

/*
 * sets up a sockets, binds the socket, and listens for incoming connections
 */
void awaitConnections() {

	int len, listen_fd, con_fd;
	struct sockaddr_in s_server, s_client;
	char *buff; /* buffer */

	/* 
	 * we have one buffer (5 B large) for our connections, so simulataneous
	 * connections are not possible here
	 */
	buff = malloc(sizeof(char)*5);

	/*
	 * sets up a socket which is just an endpoint for communication
	 */
	if( (listen_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0 ) {
		perror("socket");
		exit(1);
	}

	/*
	 *  the point of the following code is to fill the struct s_server
	 *  with our server information.  When we use bind(), we associate
	 *  a socket with parameters in s_server.  Parameters include addresses
	 *  of machines we want to allow, the port we will listen on, the
	 *  protocol we'll be using, etc.
	 */
	bzero(&s_server, sizeof(&s_server));
	memset(&s_server.sin_zero, '\0', 8);

	s_server.sin_family = AF_INET;
	s_server.sin_port = htons(port);
	s_server.sin_addr.s_addr = htons(INADDR_ANY);

	/*
	 *  here, we associate the s_server with our socket
	 *  on failure, we close the socket with shutdown()
	 */
	if( bind(listen_fd, (struct sockaddr *)&s_server, 
				sizeof(struct sockaddr)) < 0) {
		if( shutdown(listen_fd, 2) < 0)
			perror("shutdown");
		perror("bind");
		exit(1);
	}

	/*
	 * We want to notify the system that we'll be waiting for connections
	 * on this socket.  If listen() is called, we can use accept() to get
	 * the next incoming connection.  The second parameter to listen()
	 * indicates the size of the queue that the system will keep for us
	 * for incoming connections
	 */
	if( listen(listen_fd, 2) < 0 ) {
		perror("listen");
		exit(1);
	}


	/* we'll loop indefinitely for new incoming connections */
	while(1) {
		len = sizeof(s_client);

		/* 
		 * accept() associated a descriptor with the next incoming
		 * connection.  Information about the client who is connecting
		 * to us is filled in s_client.  After accept(), we can use
		 * the descriptor to send and receive to/from the client
		 */
		if( (con_fd = accept(listen_fd, (struct sockaddr *)&s_client, 
						(socklen_t *)&len) ) < 0) {
			perror("accept");
			exit(1);
		}

		/*
		 * recv is like read(), but for the descriptor for the connection,
		 * not for a file descriptor.  We read a max of 5B because that'
		 * all we've allocated.
		 */
		if( recv(con_fd, buff, sizeof(char)*5, 0) < 0) {
			perror("recv");
			exit(1);
		}

		/* what did we receive ? */
		printf(" server received %s\n", buff);

		/* this descriptor is used for the recv, we're done with it */
		close(con_fd);  /* we close the descriptor */

	}

	free(buff);

	exit(1);
}

/*
 * send a message to ourselves
 */
void sendMessage() {
	int sock_out;
	struct sockaddr_in server_addr;

	sock_out = socket(AF_INET, SOCK_STREAM, 0);

	bzero(&server_addr, sizeof(server_addr));
	server_addr.sin_family = AF_INET;
	server_addr.sin_port = htons(port);
	server_addr.sin_addr.s_addr = htonl(INADDR_ANY);

	if(connect(sock_out, (struct sockaddr *)&server_addr, 
				sizeof(struct sockaddr)) < 0 ) {
		perror("connect");
		printf("exiting..");
		if( shutdown(sock_out, 2) < 0)
			perror("shutdown");
		cleanup();
		exit(1);
		}

	if( send(sock_out, "hi!", strlen("hi!")+1, 0) < 0 ) {
		perror("send");
		exit(1);
	}

	shutdown(sock_out, 2);
}


void cleanup() {

}
